/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* compress.c

   Compresses graph by replacing instances of a substructure by a pointer
   to the substructure definition.

*/

#include "subdue.h"


graphptr compress(g, sub, numsub)
     graphptr g;
     substructure *sub;
     shortint numsub;
/* Compresses instances of sub (numbered numsub) in g. */
{
  graphptr ng = (graphptr) malloc(sizeof(graph));
  graph_list_node_ptr glnptr;
  shortint i, vmax, vnext, label, oldnv = g->nv, oldne = g->ne;
  char str[TOKENLEN];

  if (Trace > 2) {
    printf("\nEntering compress with numsub = %u and:\n",numsub);
    print_graph(g); printf("\n");
    print_substructure(sub); printf("\n");
  }
  ng->nv = ng->ne = 0;           /* Make a new graph */
  ng->refcount = 1;
  ng->vertices = (vptr *) malloc(oldnv*sizeof(vptr));
  ng->edges    = (eptr *) malloc(oldne*sizeof(eptr));
  for (i=0; i<oldnv; i++)
    ng->vertices[i] = (vptr) malloc(sizeof(vertex));
  for (i=0; i<oldne; i++)
    ng->edges[i] = (eptr) malloc(sizeof(edge));
  ng->matchcost = 0.0;
  vmax = newvertnum(g);
  vnext = vmax;
  sprintf(str, "sub %d", numsub);
  label = enter_label(str);
  glnptr = sub->instances;       /* Add substructure instances to new graph */
  while (glnptr != NULL) {       /* Make one vertex for each instance */
    ng->nv++;
    if (ng->nv > oldnv) {
      ng->vertices = (vptr *) realloc(ng->vertices, ng->nv*sizeof(vptr));
      ng->vertices[ng->nv - 1] = (vptr) malloc(sizeof(vertex));
      oldnv = ng->nv;
    }
    ng->vertices[ng->nv - 1]->number = vnext++;
    ng->vertices[ng->nv - 1]->label = label;
    if (Overlap)
      oldne = find_overlaps(ng, glnptr, sub->instances, oldne);
    glnptr = glnptr->next;
  }
  connect(g, ng, sub, vmax, oldnv, oldne);  /* Add the rest of the graph */
  if (Trace > 2) {
    printf("\nExiting compress with graph:\n");
    print_graph(ng); printf("\n");
  }
  return(ng);
}
 

void connect(g, ng, sub, subnum, oldnv, oldne)
     graphptr g, ng;
     substructure *sub;
     shortint subnum, oldnv, oldne;
/* Copy to graph ng the vertices and edges of g that are not in the
   instances of substructure sub.  Vertex numbers of replaced instances
   start at subnum. */
{
  boolean found;
  shortint i, j, from, to, subn;
  graphptr tempg;
  graph_list_node_ptr glnptr;

  for (i=0; i<g->nv; i++) {
    if (! vertex_in_sub(g->vertices[i]->number, sub)) {    /* not in */
      ng->nv++;
      if (ng->nv > oldnv) {
	ng->vertices = (vptr *) realloc(ng->vertices, ng->nv*sizeof(vptr));
	ng->vertices[ng->nv - 1] = (vptr) malloc(sizeof(vertex));
	oldnv = ng->nv;
      }
      /* Copy vertex over from old graph */
      ng->vertices[ng->nv - 1]->number = g->vertices[i]->number;
      ng->vertices[ng->nv - 1]->label  = g->vertices[i]->label;
    }
  }
  for (i=0; i<g->ne; i++) {   /* Draw edges between substructure and graph */
    if (! edge_in_sub(g->edges[i], sub)) {   /* not completely in sub */
      ng->ne++;
      if (ng->ne > oldne) {
	ng->edges = (eptr *) realloc(ng->edges, ng->ne*sizeof(eptr));
	ng->edges[ng->ne - 1] = (eptr) malloc(sizeof(edge));
	oldne = ng->ne;
      }
      from = g->edges[i]->from;
      to   = g->edges[i]->to;
      ng->edges[ng->ne - 1]->from  = from;
      ng->edges[ng->ne - 1]->to    = to;
      ng->edges[ng->ne - 1]->label = g->edges[i]->label;
      ng->edges[ng->ne - 1]->directed = g->edges[i]->directed;
      subn = subnum;
      glnptr = sub->instances;
      while (glnptr != NULL) {    /* Scan through all substructure instances */
	tempg = glnptr->g;
	for (j=0; j<tempg->nv; j++) {
	  if (tempg->vertices[j]->number == from)
	    ng->edges[ng->ne - 1]->from = subn;
	  if (tempg->vertices[j]->number == to)
	    ng->edges[ng->ne - 1]->to = subn;
	}
	glnptr = glnptr->next;
	subn++;
      }
    }
  }
}


boolean edge_in_sub(e,sub)
     eptr e;
     substructure *sub;
/* Returns TRUE if edge e is in an instance of substructure sub. */
{
  shortint i;
  graphptr g;
  graph_list_node_ptr glnptr;
  
  glnptr = sub->instances;
  while (glnptr != NULL) {
    g = glnptr->g;
    for (i=0; i<g->ne; i++)
      if (g->edges[i] == e)
	return(TRUE);
    glnptr = glnptr->next;
  }
  return(FALSE);
}


boolean vertex_in_sub(num, sub)
     shortint num;
     substructure *sub;
/* Returns TRUE if vertex num is in an instance of substructure sub. */
{
  shortint i;
  graphptr g;
  graph_list_node_ptr glnptr;

  glnptr = sub->instances;
  while (glnptr != NULL) {
    g = glnptr->g;
    for (i=0; i<g->nv; i++)
      if (g->vertices[i]->number == num)
	return(TRUE);
    glnptr = glnptr->next;
  }
  return(FALSE);
}


shortint find_overlaps(ng, glnptr, insts, oldne)
     graphptr ng;
     graph_list_node_ptr glnptr, insts;
     shortint oldne;
{
  shortint i, subnum = 0;
  graphptr instg = glnptr->g;
  char str[TOKENLEN];

  while (insts != glnptr) {
    for (i=0; i<insts->g->nv; i++) {
      if (share_vertex(instg, insts->g->vertices[i]->number)) {
	sprintf(str, "s-v %d", insts->g->vertices[i]->number);
	oldne = add_share(ng, ng->vertices[subnum]->number, str, oldne);
      }
    }
    for (i=0; i<insts->g->ne; i++) {
      if (share_edge(instg, insts->g->edges[i])) {
	sprintf(str, "s-e %d/%d", insts->g->edges[i]->from,
		insts->g->edges[i]->to);
	oldne = add_share(ng, ng->vertices[subnum]->number, str, oldne);
      }
    }
    insts = insts->next;
    subnum++;
  }
  return(oldne);
}


boolean share_vertex(g, num)
     graphptr g;
     shortint num;
/* Returns TRUE if vertex num is part of graph g. */
{
  shortint i;
  for (i=0; i<g->nv; i++)
    if (g->vertices[i]->number == num)
      return(TRUE);
  return(FALSE);
}


boolean share_edge(g, e)
     graphptr g;
     eptr e;
/* Returns TRUE if edge e is in graph g according to the from and to
   vertex numbers. */
{
  shortint i;
  for (i=0; i<g->ne; i++)
    if ((g->edges[i]->from == e->from) &&
	(g->edges[i]->to == e->to))
      return(TRUE);
  return(FALSE);
}


shortint newvertnum(g)
     graphptr g;
/* Returns one more than the maximum vertex number used in g so that new
   vertices (representing substructures) will have new, unique numbers. */
{
  shortint i, max = 0;
  for (i=0; i<g->nv; i++)
    if (g->vertices[i]->number > max)
      max = g->vertices[i]->number;
  return(max+1);
}


shortint add_share(ng, from, label, oldne)
     graphptr ng;
     shortint from, oldne;
     char *label;
/* Add edge between two substructure nodes indicating that they share part of
   the substructure (a node or an edge) */
{
  ng->ne++;
  if (ng->ne > oldne) {
    ng->edges = (eptr *) realloc(ng->edges, ng->ne*sizeof(eptr));
    ng->edges[ng->ne - 1] = (eptr) malloc(sizeof(edge));
    oldne = ng->ne;
  }
  ng->edges[ng->ne - 1]->from = from;
  ng->edges[ng->ne - 1]->to = ng->vertices[ng->nv - 1]->number;
  ng->edges[ng->ne - 1]->label = enter_label(label);
  ng->edges[ng->ne - 1]->directed = FALSE;
  return(oldne);
}
