/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* mdl.c

MDL encoding of a graph.  See JAIR paper for explanation of encoding.

*/

#include "subdue.h"


double mdl(g)
     graphptr g;
/* Returns the number of bits in the description length of the graph g.
   See relevant papers for explanation of encoding.  Note that we use
   NumLabels as the number of unique labels in the original graph, but
   this may change as the compress function adds new labels to the
   global label list. */
{
  shortint i, k, m, maxk, ones = 0;
  double vbits = 0.0, rbits = 0.0, ebits = 0.0;

  if (Trace > 2) {
    printf("\nEntering mdl with graph:\n");
    print_graph(g); printf("\n");
  }
  if (g->nv == 1) {
    if (Trace > 2) printf("Exiting mdl with value 0.0\n");
    return((double) 0.0);
  }
  /* compute bits vbits needed for vertices */
  vbits = log2((double) g->nv) + (g->nv * log2((double) NumLabels));
  /* compute bits rbits needed for rows of adjacency matrix */
  maxk = 0;
  for (i=0; i<g->nv; i++) {
    k = row_edge(g,i);
    if (k > maxk) maxk = k;
    ones += k;
    rbits += log2comb(g->nv, k);
  }
  rbits = (g->nv + 1) * log2((double) (maxk + 1)) * rbits;
  /* compute bits ebits for edges */
  m = max_edge(g);
  ebits = (g->ne * (log2((double) NumLabels) + 1)) +
          ((ones + 1)*log2((double) m));
  if (Trace > 2) printf("Exiting mdl with value %f\n",(vbits+rbits+ebits));
  return(vbits + rbits + ebits);
}


shortint row_edge(g, row)
   graphptr g;
   shortint row;
/* Return number of digit 1's in the given row of the adjacency matrix  */
{
  boolean found;
  shortint i, j, count = 0, from;
  shortint *to = (shortint *) malloc(g->nv * sizeof(shortint));

  from = g->vertices[row]->number;
  for (i=0;i<g->ne;i++) {
    found = FALSE;
    if (from == g->edges[i]->from) {
      for (j=0;j<count;j++)
	if (to[j] == g->edges[i]->to)
	  found = TRUE;
      if (! found) {
	to[count]=g->edges[i]->to;
	count++;
      }
    }
  }
  free(to);
  return count;
}


shortint max_edge(g)
   graphptr g;
/* Return maximum number of edges which have the same from node & to node */
{
  shortint i, j, n, m=0;
  shortint from, from1, to, to1;

  for (i=0;i<g->ne;i++) {
    from = g->edges[i]->from; 
    to = g->edges[i]->to;
    n = 1;
    for (j=0;j<g->ne;j++) 
      if (i != j) {
	from1 = g->edges[j]->from;
	to1 = g->edges[j]->to;
	if ((from == from1) && (to == to1)) n++;
      }
    if (n > m) m = n;
  }
  return m;
}


double log2comb(n, k)
   shortint n, k;
/* Returns log2 of combinations of n things chosen k at a time. */ 
{
  return(log2nfact(n) - log2nfact(k) - log2nfact(n-k));
}


double log2nfact(n)
   shortint n;
/* log2 of factorial using upper bound on Stirling's Approximation:
       n! <= sqrt(2*pi*n)*(n/e)^(n+1/12n) */
{
  if (n == 0)
    return((double) 1.0);
  return((0.5 * (log2((double) (PI2*n)))) +
	 ((n + (1.0/(12.0*n))) * log2((double) (n / exp((double) 1.0)))));
}


double log2(a)
   double a;
{
  if (a == 0.0)
    return((double) 0.0);
  else return(log(a)/log((double) 2.0));
}
