/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* subdue.c

   Top level SUBDUE functions and initialization.
*/

#include "subdue.h"


sub_list_node_ptr subdue(g,limit)
     graphptr g;
     shortint limit;
/* Initialize Limit and call main discovery procedure.  If limit=0,
   then set Subdue Limit to 0.5*E(g). */
{
  sub_list_node_ptr isubs, dsubs;
  Graph = g;
  Size = Graph->nv + Graph->ne;  
  if (limit != 0) Limit = limit;
  else {
    Limit = (shortint) ceil((Graph->ne / 2.0));
    if (Trace > 0) printf("Limit set to %u\n",Limit);
  }
  isubs = initial_subs();  /* freed by discover_subs */
  dsubs = discover_subs(isubs);
  return(dsubs);
}


sub_list_node_ptr initial_subs()
/* Returns a list of substructures consisting of one vertex.
   There will be one substructure for each unique vertex label
   in the global Graph. */
{
  sub_list_node_ptr subs = NULL;
  substructure *isub;
  shortint nv = Graph->nv;
  vptr *vptrs = Graph->vertices;
  shortint i;

  for (i = 0; i < nv; i++, vptrs++)
    if (! label_member(*vptrs,subs)) {
      isub = initial_substructure(*vptrs);
      subs = insert_sub(isub,subs,FALSE);
    }
  return subs;
}


boolean label_member(v, subs)
     vptr v;
     sub_list_node_ptr subs;
/* Returns TRUE if the label in vertex v is already in a vertex
   on the list of substructures subs. */
{
  boolean member = FALSE;
  sub_list_node_ptr slnptr;

  slnptr = subs;
  while ((! member) && (slnptr != NULL)) {
    if (v->label == slnptr->s->definition->vertices[0]->label)
      member = TRUE;
    slnptr = slnptr->next;
  }
  return member;
}


substructure *initial_substructure(v)
     vptr v;
/* Returns a substructure whose definition corresponds to the
   given vertex v, and whose instances are the vertices in
   the Graph having the same label as v.  */
{
  substructure *subptr;
  vptr *vdef;

  vdef = (vptr *) malloc(sizeof(vptr));
  *vdef = v;
  subptr = (substructure *) malloc(sizeof(substructure));
  subptr->value = 0.0;
  subptr->definition = build_graph(1,0,vdef,NULL,0.0);
  subptr->instances = get_instances(v);
  evaluate_sub(subptr);
  return subptr;
}


graph_list_node_ptr get_instances(v)
     vptr v;
/* Returns a list of graph instances, one for each vertex in Graph having
   the same label as v. */
{
  graph_list_node_ptr instances = NULL;
  graph_list_node_ptr glnptr;
  vptr *gv = Graph->vertices;
  vptr *vs;
  shortint i, nv;

  nv = Graph->nv;
  for (i = 0; i < nv; i++, gv++) {
    if (v->label == gv[0]->label) {
      vs = (vptr *) malloc(sizeof(vptr));
      *vs = *gv;
      glnptr = (graph_list_node_ptr) malloc(sizeof(graph_list_node));
      glnptr->g = build_graph(1,0,vs,NULL,0.0);
      glnptr->next = instances;
      instances = glnptr;
    }
  }
  return instances;
}
