/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* util.c

   SUBDUE utility functions.
*/

#include "subdue.h"


graphptr build_graph(nv, ne, vertices, edges, matchcost)
     shortint nv, ne;
     vptr *vertices;
     eptr *edges;
     float matchcost;
{
  graphptr g;
  g = (graphptr) malloc(sizeof(graph));
  g->nv = nv;
  g->ne = ne;
  g->refcount = 1;
  g->vertices = vertices;
  g->edges = edges;
  g->matchcost = matchcost;
  return g;
}


void free_substructure(s)
     substructure *s;
{
  graph_list_node_ptr glnptr, glnptr1;

  glnptr = s->instances;
  while (glnptr != NULL) {
    glnptr->g->refcount--;
    if (glnptr->g->refcount == 0)
      free(glnptr->g);
    glnptr1 = glnptr;
    glnptr = glnptr->next;
    free(glnptr1);
  }
  free(s->definition);
  free(s);
}


shortint size(g)
     graphptr g;
{
  return(g->nv + g->ne);
}


vptr find_vertex(vn, g)
     shortint vn;
     graphptr g;
/* Returns the vertex structure corresponding to the vertex number vn
   in graph g. */
{
  vptr *vs = g->vertices;
  vptr rtnptr = NULL;
  shortint nv = g->nv;
  shortint i;
  boolean found = FALSE;

  for (i = 0; ((i < nv) && (! found)); i++) {
    if (vs[i]->number == vn) {
      found = TRUE;
      rtnptr = vs[i];
    }
  }
  return rtnptr;
}


void print_vertices(g)
     graphptr g;
{
  shortint i;
  for (i = 0; i < g->nv; i++) {
    printf("\n      %d",g->vertices[i]->number);
    printf(" (%s)",LabelList[g->vertices[i]->label]);
  }
}


void print_edges(g)
     graphptr g;
{
  shortint i;
  for (i = 0; i < g->ne; i++) {
    printf("\n      %s[%d", LabelList[g->edges[i]->label], g->edges[i]->from);
    if (g->edges[i]->directed)
      printf(" -> ");
    else printf(" -- ");
    printf("%d]",g->edges[i]->to);
  }
}


void print_graph(g, defn)
     graphptr g;
     boolean defn;
/* If defn=TRUE, then g is a definition, not an instance, so do not
   print the matchcost (because it is always 0 for a definition. */
{
  printf("  Graph: nv = %d, ne = %d",g->nv,g->ne);
  printf("\n   Vertices:");
  print_vertices(g);
  printf("\n   Edges:");
  print_edges(g);
  if (! defn) printf("\n   MatchCost = %f",g->matchcost);
  printf("\n");
}


void print_substructure(s)
     substructure *s;
/* Prints the substructure graph definition.  If Trace > 1,
   the substructure instances are also printed. */
{
  shortint count = 0;
  graph_list_node_ptr glnptr = s->instances;

  printf("Substructure: value = %f\n",s->value);
  printf(" Definition:\n");
  print_graph(s->definition,TRUE);
  if (Trace > 1) printf(" Instances:\n");
  while (glnptr != NULL) {
    count++;
    if (Trace > 1) {
      printf("  (%u)",count);
      print_graph(glnptr->g,FALSE);
    }
    glnptr = glnptr->next;
  }
  if (Trace <= 1) printf(" %u instances.\n",count);
}


void print_sub_list(subs)
     sub_list_node_ptr subs;
{
  while (subs != NULL) {
    print_substructure(subs->s);
    printf("\n");
    subs = subs->next;
  }
}
